from .common import KibanaApiModelBase, KibanaApiBase
import uuid
from functools import reduce
from .paneltools import append_panel


class Dashboard(KibanaApiModelBase):
    """
    Kibana Dashboard model
    """
    id = None
    title = None
    hits = 0
    description = None
    panels = None
    options = None
    ui_state = None
    version = 1
    time_restore = True
    time_to = "now"
    time_from = "now-7d"
    refresh_interval_value = 0
    search_source = None

    @classmethod
    def from_kibana(cls, es_document):
        if es_document.get('_type') != 'dashboard':
            raise Exception('The document is not a dashboard')
        source = es_document['_source']
        d = Dashboard()
        d.id = es_document['_id']
        d.title = source['title']
        d.description = source['description']
        d.hits = source['hits']
        d.panels = cls.unserialize(source['panelsJSON'])
        d.options = cls.unserialize(source['optionsJSON'])
        d.ui_state = cls.unserialize(source['uiStateJSON'])
        d.version = source['version']
        d.time_restore = source['timeRestore']
        d.time_to = source['timeTo']
        d.time_from = source['timeFrom']
        d.refresh_interval_value = source['refreshInterval']['value']
        d.search_source = cls.unserialize(source['kibanaSavedObjectMeta']['searchSourceJSON'])
        return d

    def to_kibana(self):
        return {
            'title': self.title,
            'description': self.description,
            'hits': self.hits,
            'panelsJSON': self.serialize(self.panels),
            'optionsJSON': self.serialize(self.options),
            'uiStateJSON': self.serialize(self.ui_state),
            'version': self.version,
            'timeRestore': self.time_restore,
            'timeTo': self.time_to,
            'timeFrom': self.time_from,
            'refreshInterval': {
                'display': "Off",
                'pause': False,
                'value': self.refresh_interval_value
            },
            'kibanaSavedObjectMeta': {
                'searchSourceJSON': self.serialize(self.search_source)
            }
        }

    def __unicode__(self):
        return self.title

    def add_visualization(self, visualization, size_x=6, size_y=3, col=0, row=0):
        """
        Adds the visualization to the dashboard. Leave col and row = 0 for automatic placement of the visualization.
        Visualizations are placed on a grid with 12 columns and unlimited rows.
        :param visualization: previously loaded visualization
        :param size_x width of the panel
        :param size_y height of the panel
        :param col 1-based column of the top left corner, leave 0 for automatic placement
        :param row 1-based row of the top left corner, leave 0 for automatic placement
        :return: newly created panel or None
        """
        new_panel_index = self.get_max_index()+1
        if col and row:
            new_panel = {
                'col': col, 'row': row,
                'size_x': size_x, 'size_y': size_y,
                'panelIndex': new_panel_index,
                'type': 'visualization',
                'id': visualization.id
            }
            self.panels.append(new_panel)
            return new_panel
        else:
            new_panel = append_panel(self.panels, size_x, size_y)
            if new_panel:
                new_panel['id'] = visualization.id
                new_panel['panelIndex'] = new_panel_index
                new_panel['type'] = 'visualization'
                return new_panel

    def remove_visualization(self, visualization_id):
        """
        Removes all visualizations with the specified id from the dashboard
        :param visualization_id:
        :return:
        """
        for i, panel in enumerate(self.panels):
            if panel['id'] == visualization_id:
                del self.panels[i]

    def get_max_index(self):
        return reduce(lambda max_index, panel: max(max_index, panel['panelIndex']), self.panels, 0)


class DashboardsManager(KibanaApiBase):
    """
    Manages Kibana dashboards
    """
    index = '.kibana'
    doc_type = 'dashboard'

    def get_all(self):
        """
        Returns a list of all dashboards
        :return:
        """
        res = self.es.search(index=self.index, doc_type=self.doc_type, body={'query': {'match_all': {}}})
        if not res['hits']['total']:
            return []
        return [Dashboard.from_kibana(hit) for hit in res['hits']['hits']]

    def add(self, dashboard):
        """
        Creates a new dashboard
        :param dashboard: instance of Dashboard
        :return:
        """
        res = self.es.create(index=self.index, id=str(uuid.uuid1()), doc_type=self.doc_type,
                             body=dashboard.to_kibana(), refresh=True)

    def update(self, dashboard):
        """
        Updates the existing dashboard
        :param dashboard: instance of Dashboard that was previously loaded
        :return:
        """
        res = self.es.update(index=self.index, id=dashboard.id, doc_type=self.doc_type,
                             body={'doc': dashboard.to_kibana()},
                             refresh=True)
        return res

    def remove(self, dashboard):
        """
        Deletes the dashboard
        :param dashboard:
        :return:
        """
        res = self.es.delete(index=self.index, id=dashboard.id, doc_type=self.doc_type, refresh=True)
        return res




import json


class KibanaApiBase(object):
    es = None

    def __init__(self, elastic_search):
        self.es = elastic_search


class KibanaApiModelBase(object):
    @classmethod
    def serialize(cls, data):
        return json.dumps(data, ensure_ascii=False)

    @classmethod
    def unserialize(cls, string):
        return json.loads(string)



from .common import KibanaApiBase, KibanaApiModelBase
import uuid


class Visualization(KibanaApiModelBase):
    """
    Visualization model
    """
    id = None
    title = None
    description = None
    vis_state = None
    ui_state = None
    search_source = None

    @classmethod
    def from_kibana(cls, es_document):
        if es_document.get('_type') != 'visualization':
            raise Exception('The document is not visualization')
        source = es_document['_source']
        v = Visualization()
        v.id = es_document['_id']
        v.title = source['title']
        v.description = source['description']
        v.ui_state = source['uiStateJSON']
        v.vis_state = cls.unserialize(source['visState'])
        v.search_source = cls.unserialize(source['kibanaSavedObjectMeta']['searchSourceJSON'])
        return v

    def to_kibana(self):
        return {
            'title': self.title,
            'description': self.description,
            'visState': self.serialize(self.vis_state),
            'uiStateJSON': self.ui_state,
            'kibanaSavedObjectMeta': {
                'searchSourceJSON': self.serialize(self.search_source)
            }
        }

    def __str__(self):
        return self.title

    def url_path(self):
        return "/app/kibana#/visualize/edit/{id}?embed=true&_g=(refreshInterval%3A(display%3AOff%2Cpause%3A!f%2Cvalue%3A0)%2Ctime%3A(from%3Anow-7d%2Cmode%3Aquick%2Cto%3Anow))".format(
            id=self.id)


class VisualizationsManager(KibanaApiBase):
    """
    Manages Kibana visualizations
    """
    index = '.kibana'
    doc_type = 'visualization'

    def get_all(self):
        """
        Returns a list of all visualizations
        :return: list of the Visualization instances
        """
        res = self.es.search(index=self.index, doc_type=self.doc_type, body={'query': {'match_all': {}}})
        if not res['hits']['total']:
            return []
        return [Visualization.from_kibana(hit) for hit in res['hits']['hits']]

    def add(self, visualization):
        """
        Creates a new visualization
        :param visualization: instance of Visualization
        :return:
        """
        res = self.es.create(index=self.index, id=str(uuid.uuid1()), doc_type=self.doc_type,
                             body=visualization.to_kibana(), refresh=True)
        return res

    def update(self, visualization):
        """
        Updates existing visualization
        :param visualization: instance of Visualization that was previously loaded
        :return:
        """
        res = self.es.update(index=self.index, id=visualization.id, doc_type=self.doc_type,
                             body={'doc': visualization.to_kibana()},
                             refresh=True)
        return res

    def remove(self, visualization):
        """
        Deletes the visualization
        :param visualization: instance of Visualization that was previously loaded
        :return:
        """
        res = self.es.delete(index=self.index, id=visualization.id, doc_type=self.doc_type, refresh=True)
        return res


class VisualizationTemplatesManager(VisualizationsManager):
    """
    Manages visualization templates. Creates a new document type vis_template in the Kibana index
    """
    index = '.kibana'
    doc_type = 'vis_template'



def bottoms(panels):
    """
    Finds bottom lines of all panels
    :param panels:
    :return: sorted by row list of tuples representing lines (col, row , col + len, row)
    """
    bottom_lines = [(p['col'], p['row'] + p['size_y'], p['col'] + p['size_x'], p['row'] + p['size_y']) for p in panels]
    return sorted(bottom_lines, key=lambda l: l[1], reverse=True)


def find_shape(bottom_lines, max_len):
    """
    Finds a shape of lowest horizontal lines with step=1
    :param bottom_lines:
    :param max_len:
    :return: list of levels (row values), list indexes are columns
    """
    shape = [1] * max_len
    for i in range(max_len):
        for line in bottom_lines:
            if line[0] <= i + 1 < line[2]:
                shape[i] = line[1]
                break
    return shape


def longest_lines(shape):
    """
    Creates lines from shape
    :param shape:
    :return: list of dictionaries with col,row,len fields
    """
    lines = []
    for level in set(shape):
        count = 0
        for i in range(len(shape)):
            if shape[i] <= level:
                count += 1
            elif count:
                lines.append({'row': level, 'col': i - count + 1, 'len': count})
                count = 0
        if count:
            lines.append({'row': level, 'col': i - count + 2, 'len': count})

    return sorted(lines, key=lambda l: l['row'])


def find_place(lines, size_x):
    """
    Finds the highest place at the left for a panel with size_x
    :param lines:
    :param size_x:
    :return: line with row, col, len
    """
    for line in lines:
        if line['len'] >= size_x:
            return line


def append_panel(panels, size_x, size_y, max_col=12):
    """
    Appends a panel to the list of panels. Finds the highest palce at the left for the new panel.
    :param panels:
    :param size_x:
    :param size_y:
    :param max_col:
    :return: a new panel or None if it is not possible to place a panel with such size_x
    """
    bottom_lines = bottoms(panels)
    shape = find_shape(bottom_lines, max_col)
    lines = longest_lines(shape)
    line = find_place(lines, size_x)
    if not line:
        return
    panel = {
        'col': line['col'],
        'row': line['row'],
        'size_x': size_x,
        'size_y': size_y,
    }
    panels.append(panel)
    return panel



from setuptools import setup

setup(name='kibana_dashboard_api',
      version='0.1.0',
      description='Manages Kiabana visualizations and dashboards.',
      long_description='List/create/delete Kibana visualizations and dashboards manipulating Elasticsearch index.',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries'
      ],
      keywords='kibana elasticsearch',
      url='https://github.com/harabchuk/kibana-dashboard-api',
      author='Aliaksei Harabchuk',
      author_email='aliaksei.harabchuk@gmail.com',
      license='MIT',
      packages=['kibana_dashboard_api'],
      install_requires=[
          'elasticsearch',
      ],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      include_package_data=True,
      zip_safe=False)
